/*
**  Low-cost Remote Who
**  There is no copyright on this package.
**  User program to read the data files.
*/
#include "lwho.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>
#if	!defined(lint) && !defined(SABER)
static char RCS[] =
	"$Header: /nfs/papaya/source/lwho/RCS/lwho.c,v 1.4 91/03/16 18:39:35 rsalz Exp $";
#endif	/* !defined(lint) && !defined(SABER) */

typedef struct _HOSTDATA {
    int			Valid;
    int			Users;
    time_t		BootTime;
    double		LoadAve[3];
    char		Name[HOSTLEN];
} HOSTDATA;

typedef struct _USERDATA {
    int			Valid;
    time_t		Onsince;
    time_t		Idle;
    char		Name[NAMELEN];
    char		Line[LINELEN];
    HOSTDATA		*Host;
} USERDATA;


#define HOST_DELTA	96
STATIC HOSTDATA		*Hosts;

#define USER_DELTA	20
STATIC USERDATA		*Users;

int                     FakeRwho; /* Added by RKL */
int                     FakeRup; /* Added by RKL */
int                     FakeRshow; /* Added by RKL */
int                     MaxAge; /* Added by RKL */
time_t		        timeup; /* Added by RKL */
long                    curtime; /* Added by RKL */

#define DAYS_OLD        14      /* Remove files 14 days or older */
#define DOWN_HOST       300     /* Number of secs before host is considered 'down' */
#define MIN_BOOT        3600    /* Must be past 1.00 am on Jan 1st 1970 :-) */


/*
**  Get a free host entry, growing as needed.
*/
STATIC HOSTDATA *
GetHost()
{
    static HOSTDATA	*Last;
    static int		Count;
    static int		Size;

    /* Give last one back if they didn't fill it in. */
    if (Last && !Last->Valid)
	return Last;

    if (Hosts == NULL) {
	/* First time here, get space. */
	Count = 0;
	Size = HOST_DELTA;
	Hosts = NEW(HOSTDATA, Size);
    }
    else if (Count >= Size - 1) {
	/* Need more space. */
	Size += HOST_DELTA;
	Hosts = RENEW(Hosts, HOSTDATA, Size);
    }

    Hosts[Count].Valid = FALSE;
    Last = &Hosts[Count++];
    return Last;
}


/*
**  Get a free user entry, growing as needed.
*/
STATIC USERDATA *
GetUser()
{
    static USERDATA	*Last;
    static int		Count;
    static int		Size;

    /* Give last one back if they didn't fill it in. */
    if (Last && !Last->Valid)
	return Last;

    if (Users == NULL) {
	/* First time here, get space. */
	Count = 0;
	Size = USER_DELTA;
	Users = NEW(USERDATA, Size);
    }
    else if (Count >= Size - 1) {
	/* Need more space. */
	Size += USER_DELTA;
	Users = RENEW(Users, USERDATA, Size);
    }

    Users[Count].Valid = FALSE;
    Last = &Users[Count++];
    return Last;
}


/*
**  Sort host data by host name.
*/
STATIC int
HSalpha(p1, p2)
    char	*p1;
    char	*p2;
{
    HOSTDATA	*h1;
    HOSTDATA	*h2;

    h1 = (HOSTDATA *)p1;
    h2 = (HOSTDATA *)p2;
    return strcmp(h1->Name, h2->Name);
}


/*
**  Sort host data by time booted.
*/
STATIC int
HSboottime(p1, p2)
    char	*p1;
    char	*p2;
{
    HOSTDATA	*h1;
    HOSTDATA	*h2;
    long	l;

    h1 = (HOSTDATA *)p1;
    h2 = (HOSTDATA *)p2;
    if (l = h1->BootTime - h2->BootTime)
	return l < 0 ? -1 : 1;
    return 0;
}


/*
**  Sort host data by host load average.
*/
STATIC int
HSloadave(p1, p2)
    char	*p1;
    char	*p2;
{
    HOSTDATA	*h1;
    HOSTDATA	*h2;
    long	 l;

    h1 = (HOSTDATA *)p1;
    h2 = (HOSTDATA *)p2;
    if (l = h1->LoadAve[0] - h2->LoadAve[0])
	return l < 0 ? -1 : 1;
    return 0;
}


/*
**  Sort host data by number of users.
*/
STATIC int
HSnumusers(p1, p2)
    char	*p1;
    char	*p2;
{
    HOSTDATA	*h1;
    HOSTDATA	*h2;

    h1 = (HOSTDATA *)p1;
    h2 = (HOSTDATA *)p2;
    return h1->Users - h2->Users;
}


/*
**  Sort user data alphabetically, than by host, than by idle time.
*/
STATIC int
USalpha(p1, p2)
    char	*p1;
    char	*p2;
{
    USERDATA	*u1;
    USERDATA	*u2;
    int		 i;
    long	 l;
    
    u1 = (USERDATA *)p1;
    u2 = (USERDATA *)p2;
    if (i = strcmp(u1->Name, u2->Name))
	return i;
    if (i = strcmp(u1->Host->Name, u2->Host->Name))
	return i;
    if (l = u1->Idle - u2->Idle)
	return l < 0 ? -1 : 1;
    return 0;
}


/*
**  Sort user data by idle time.
*/
STATIC int
USidle(p1, p2)
    char	*p1;
    char	*p2;
{
    USERDATA	*u1;
    USERDATA	*u2;
    long	 l;

    u1 = (USERDATA *)p1;
    u2 = (USERDATA *)p2;
    if (l = u1->Idle - u2->Idle)
	return l < 0 ? -1 : 1;
    return 0;
}


/*
**  Return a formatted string with how long between a given time and now.
*/
STATIC char *
HowLong(t)
    time_t		t;
{
    static char		buff[20];
    int			hours;

    /* Get minutes and hours from seconds. */
    t = (t + 59) / 60;
    hours = t / 60;
    t -= hours * 60;

    if (FakeRup)
    {
       if (hours >= 24)
          sprintf(buff,"%4d+%02d:%02d",(int)(hours/24),
                  hours%24,t);
       else
          sprintf(buff,"%7d:%02d",hours,t);
    }
    else
    if (FakeRwho)
    {
       if (hours > 99) strcpy(buff," 99:59");
       else if ((hours == 0) && (t == hours * 60)) strcpy(buff,"");
            else if (hours == 0) sprintf(buff,"   :%02d",t);
                 else sprintf(buff," %2d:%02d",hours,t);
    }
    else
    if (FakeRshow)
    {
       if (hours > 99) strcpy(buff," 99:59");
       else if ((hours==0) && (t==0)) strcpy(buff," Busy!");
            else sprintf(buff," %02d:%02d",hours,t);
    }
    else
    if (hours >= 24)
	(void)sprintf(buff, "%3dd ", hours / 24);
    else
	(void)sprintf(buff, "%2d:%02d", hours, t);
    return buff;
}


/*
**  Split the cruft out of a ctime() result.
*/
STATIC char *
Myctime(t)
    time_t	*t;
{
    char	*p;

    p = ctime(t);
    if (FakeRwho) p[16] = '\0'; else p[19] = '\0';
    return p + 4;
}


/*
**  Do luptime output.
*/
STATIC void
ShowHostData(Sorter, Reversed, Banner)
    int			Sorter;
    int			Reversed;
    int			Banner;
{
    register HOSTDATA	*H;
    register int	i;
    time_t		now;

    /* Count those pups. */
    for (H = Hosts, i = 0; H->Valid; H++, i++)
	;
    if (i == 0)
	return;

    /* Sort that sucker. */
    switch (Sorter) {
    default:
    case 'a':
    case 'i':
	qsort((char *)Hosts, i, sizeof *Hosts, HSalpha);
	break;
    case 'b':
	qsort((char *)Hosts, i, sizeof *Hosts, HSboottime);
	break;
    case 'l':
	qsort((char *)Hosts, i, sizeof *Hosts, HSloadave);
	break;
    case 'n':
	qsort((char *)Hosts, i, sizeof *Hosts, HSnumusers);
	break;
    }

    if (Banner && !FakeRup)
	printf("%*s Users    Load Average      Up       Booted\n",
	    -HOSTLEN / 2, "Host");

    /* Watch the clock. */
    now = time((time_t *)NULL);

    /* Dump the data. */
    if (Reversed) {
	Reversed = -1;
	H = &Hosts[i - 1];
    }
    else {
	Reversed = 1;
	H = &Hosts[0];
    }
    for ( ; --i >= 0; H += Reversed)
    {
        timeup=(time_t)(H->BootTime);
        if ((timeup>MIN_BOOT) || (timeup<0))
           if (FakeRup)
           {
              if (timeup<0) { printf("%-12sdown%s\n",H->Name, HowLong(-timeup)); }
              else
              { 
                 printf("%-14sup%s, %3d user", H->Name, HowLong(now - timeup), H->Users);
                 if (H->Users != 1) printf("s,"); else printf(", ");
    	         printf("  load %5.2f %5.2f %5.2f\n", H->LoadAve[0], H->LoadAve[1], H->LoadAve[2]);
              }
           }
           else
	    printf("%*s %3d   %5.2f %5.2f %5.2f  %s  %s\n", -HOSTLEN / 2, H->Name,
	    H->Users, H->LoadAve[0], H->LoadAve[1], H->LoadAve[2],
	    HowLong(now - timeup), Myctime(&timeup));
    }
}


/*
**  Do lwho output.
*/
STATIC void
ShowUserData(Sorter, Reversed, Banner)
    int			Sorter;
    int			Reversed;
    int			Banner;
{
    register USERDATA	*U;
    register int	i,loop,diff;
    long                laziest,earliest;
    time_t              now;
    char *fullterm = "This holds the tty";
    char olduser[16],dummy[64],realname[64],phone[32],lasthost[32];
    char machlist[255];
    struct passwd *passnull=NULL;
    struct passwd *passinfo;

    /* Count those pups. */
    for (U = Users, i = 0; U->Valid; U++, i++)
	;
    if (i == 0)
	return;

    /* Watch the clock. */
    now = time((time_t *)NULL);
    /* Sort that sucker. */
    switch (Sorter) {
    default:
    case 'a':
    case 'b':
    case 'l':
    case 'n':
	qsort((char *)Users, i, sizeof *Users, USalpha);
	break;
    case 'i':
	qsort((char *)Users, i, sizeof *Users, USidle);
	break;
    }

    if (Banner)
       if (FakeRshow)
        printf("%*s Real Life Name          On     Idle    Phone  Machines\n",
            -NAMELEN, "Username");
       else
       if (!FakeRwho)
	printf("%*s %*s %*s %s %s\n",
	    -NAMELEN, "Name",
	    -HOSTLEN / 2-3, "Host",
	    -LINELEN+2, "TTY",
	    "Idle", "    On since");

    /* Dump the data. */
    if (Reversed) {
	Reversed = -1;
	U = &Users[i - 1];
    }
    else {
	Reversed = 1;
	U = &Users[0];
    }
    strcpy(olduser,U->Name);
    laziest=U->Idle;
    earliest=(time_t)(U->Onsince);
    if (earliest<=MIN_BOOT) earliest=now;
    strcpy(lasthost,U->Host->Name);
    strcpy(machlist,lasthost);
    for ( ; --i >= 0; U += Reversed)
    {
        timeup=(time_t)(U->Onsince);
        if ((U->Name[0]!='\0') && (timeup<0 || timeup>MIN_BOOT))
        {
           loop=0;
           while ((U->Line[loop]!='\0') &&
                  (U->Line[loop]!=':') &&
                  (U->Line[loop]!='.')) loop++;
           U->Line[loop]='\0';
           if (FakeRwho || FakeRshow)
           {
              if (fullterm=strrchr(U->Line,'/')) fullterm++;
                 else fullterm=U->Line;
              if (FakeRwho)
              printf("%-9s%s:%*s%-10s%s\n",U->Name,U->Host->Name,
                     strlen(U->Host->Name)-17,fullterm,
                     Myctime(&timeup),HowLong(U->Idle));
              else
              {
              while (diff=(strcmp(U->Name,olduser)!=0) || i==0)
              {
                 if (i==0 && !diff)
                 {
                    if (U->Idle<laziest) laziest=U->Idle;
                    if (timeup<earliest && timeup>MIN_BOOT) earliest=timeup;
                    if (strcmp(U->Host->Name,lasthost)!=0)
                    {
                       strcpy(lasthost,U->Host->Name);
                       strcat(machlist,",");
                       strcat(machlist,lasthost);
                    }
                 }
                 realname[0]='\0'; phone[0]='\0';
                 passinfo=getpwnam(olduser);
                 if (passinfo!=passnull)
                    sscanf(passinfo->pw_gecos,"%[^,],%[^,],%[^,]",
                           realname,dummy,phone);
                 if (realname[0]=='\0') strcpy(realname,"Unknown");
                 if (phone[0]=='\0' || phone[0]==' ')
                    strcpy(phone,"N/A");
                 if (strlen(realname)>20)
                 {
                    realname[18]='.'; realname[19]='.'; realname[20]='\0';
                 }
                 strcpy(dummy,HowLong(now-earliest));
                 printf("%-11s%-22s%-8s%-9s%-7s%s\n",
                        olduser,realname,dummy,HowLong(laziest),phone,machlist);
                 if (i)
                 {
                    strcpy(olduser,U->Name);
                    laziest=U->Idle;
                    if (timeup>MIN_BOOT) earliest=timeup; else earliest=now;
                    strcpy(lasthost,U->Host->Name);
                    strcpy(machlist,lasthost);
                 }
                 else i--;
              }
              if (U->Idle<laziest) laziest=U->Idle;
              if (timeup<earliest && timeup>MIN_BOOT) earliest=timeup;
              if (strcmp(U->Host->Name,lasthost)!=0)
              {
                 strcpy(lasthost,U->Host->Name);
                 strcat(machlist,",");
                 strcat(machlist,lasthost);
              }
              }
           }
           else
           printf("%*s %*s %*s %s  %s\n",
                  -NAMELEN, U->Name,
                  -HOSTLEN / 2, U->Host->Name,
                  -LINELEN, U->Line,
                  HowLong(U->Idle), Myctime(&timeup));
        }
    }
}


/*
**  Return TRUE if a string is a member of a NULL-terminated array.
*/
STATIC int
Found(argv, p)
    register char	**argv;
    register char	*p;
{
    for ( ; *argv; argv++)
	if (**argv == *p && strcmp(*argv, p) == 0)
	    return TRUE;
    return FALSE;
}


main(ac, av)
    int			ac;
    char		*av[];
{
    register HOSTDATA	*H;
    register USERDATA	*U;
    register DIR	*dp;
    register FILE	*F;
    register int	i;
    struct dirent	*ep;
    char		buff[BUFFLEN];
    char		**Ihosts;
    char		**Iusers;
    int			Ihostcount;
    int			Iusercount;
    int			DoHosts;
    int			Reversed;
    int			SortMode;
    int			Banner;
    long		l[3];
    struct stat         filestats; /* Added by RKL */
    int                 fileage; /* Age of file in seconds */

    /* Get program name. */
    if (progname = strrchr(av[0], '/'))
	progname++;
    else
	progname = av[0];
    DoHosts = strcmp(progname, "luptime") == 0;
    FakeRwho = strcmp(progname, "rwho") == 0; /* Added by RKL */
    FakeRup = strcmp(progname, "ruptime") == 0; /* Added by RKL */
    FakeRshow = strcmp(progname, "rshow") == 0; /* Added by RKL */
    MaxAge = DAYS_OLD * 24 * 3600; /* Maximum age of file in secs */

    /* Set up defaults and arrays of "interesting" hosts/users. */
    Reversed = FALSE;
    Banner = TRUE;
    SortMode = 'a';
    Ihosts = NEW(char*, ac + 1);
    Iusers = NEW(char*, ac + 1);
    Ihostcount = 0;
    Iusercount = 0;

    /* Parse JCL. */
    while ((i = getopt(ac, av, "abilnrhpxm:u:")) != EOF)
	switch (i) {
	default:
	    exit(1);
	case 'a':
	case 'b':
	case 'i':
	case 'l':
	case 'n':
	    SortMode = i;
	    break;
	case 'r':
	    Reversed = TRUE;
	    break;
	case 'h':
	    DoHosts = TRUE;
	    break;
	case 'p':
	    DoHosts = FALSE;
	    break;
	case 'x':
	    Banner = FALSE;
	case 'm':
	    Ihosts[Ihostcount++] = optarg;
	    break;
	case 'u':
	    Iusers[Iusercount++] = optarg;
	    break;
	}
    Iusers[Iusercount] = NULL;
    Ihosts[Ihostcount] = NULL;

    /* Go to spool directory, open it. */
    if (chdir(SPOOLDIR) < 0) {
	fprintf(stderr, "Can't cd to \"%s\", %s.\n", SPOOLDIR,
	    strerror(errno));
	exit(1);
    }
    if ((dp = opendir(".")) == NULL) {
	fprintf(stderr, "Can't open directory, %s.\n", strerror(errno));
	exit(1);
    }

    /* Read data. */
    curtime = time(0); /* Get (current time - down limit) in seconds - added by RKL */
    while (ep = readdir(dp)) {
	/* Is this a file we're interested and that we can access? */
        if (ep->d_name[0]=='.') continue; /* Ignore everything beginning with a dot */

        stat(ep->d_name,&filestats);
        fileage = curtime - filestats.st_mtime;
        if (fileage>=MaxAge)
        {
            (void)unlink(ep->d_name);
            continue;
        }
        else
        if (fileage>=DOWN_HOST)
        {
           if (FakeRup) fileage= -fileage; else continue;
        }

	if ((F = fopen(ep->d_name, "r")) == NULL) continue;

	/* Check version. */
	if (fgets(buff, sizeof buff, F) == NULL
	 || strncmp(buff, VERSION, sizeof VERSION - 1) != 0) {
	    (void)fclose(F);
	    continue;
	}

	/* Is this a host the user is interested in? */
	if (Ihostcount && !Found(Ihosts, ep->d_name)) {
	    (void)fclose(F);
	    continue;
	}

	/* Get space, store the name. */
	H = GetHost();
	(void)strncpy(H->Name, ep->d_name, sizeof H->Name);
	H->Name[sizeof H->Name - 1] = '\0';

	/* Get boot time. */
	if (fgets(buff, sizeof buff, F) == NULL) {
	    (void)fclose(F);
	    continue;
	}
	if (fileage<=-DOWN_HOST) H->BootTime = fileage;
                            else H->BootTime = atol(buff);

	/* Get load average. */
	if (fgets(buff, sizeof buff, F) == NULL) {
	    (void)fclose(F);
	    continue;
	}
	if (sscanf(buff, "%ld %ld %ld", &l[0], &l[1], &l[2]) != 3) {
	    (void)fclose(F);
	    continue;
	}
	H->LoadAve[0] = l[0] / 100.;
	H->LoadAve[1] = l[1] / 100.;
	H->LoadAve[2] = l[2] / 100.;

	/* Get users. */
	for (H->Users = 0; fgets(buff, sizeof buff, F); ) {

	    /* Logical end of file? */
	    if (strncmp(buff, VERSION, sizeof VERSION - 1) == 0)
		break;

	    /* Get space, read in a line. */
	    U = GetUser();
	    if (sscanf(buff, "%s %s %ld %ld",
		    U->Name, U->Line, &U->Onsince, &U->Idle) != 4)
		continue;
	    H->Users++;

	    /* Is this a user the user is interested in? */
	    if (Iusercount && !Found(Iusers, U->Name))
		continue;

	    U->Host = H;
	    U->Valid = TRUE;
	}

	H->Valid = TRUE;

	(void)fclose(F);

    }
    (void)closedir(dp);

    /* Print data. */
    if (DoHosts || FakeRup)
	ShowHostData(SortMode, Reversed, Banner);
    else
	ShowUserData(SortMode, Reversed, Banner);

    /* That's it. */
    exit(0);
    /* NOTREACHED */
}
