/*
**  Low-cost Remote Who
**  There is no copyright on this package.
**  Deamon to write the data file.
*/
#include "lwho.h"
#include <signal.h>
#include <nlist.h>
#include <utmp.h>
#include <sgtty.h>
#include <sys/stat.h>
#ifdef	sun
#include <sys/param.h>
#endif	/* sun */
#if	!defined(lint) && !defined(SABER)
static char RCS[] =
	"$Header: /nfs/papaya/source/lwho/RCS/lwhod.c,v 1.3 91/03/16 18:29:11 rsalz Exp $";
#endif	/* !defined(lint) && !defined(SABER) */

STATIC FILE	*F;
STATIC char	 Host[BUFFLEN];
int	Kmem;
int	BootTime;
FILE	*U;
int	i;
int	stalled;
int     showerr=1;

STATIC struct nlist	Nlist[] = {
    { "avenrun" },
#define NL_AVENRUN	0
    { "boottime" },
#define NL_BOOTTIME	1
    NULL
};


/*
**  We're going down:  close open file, remove our data file.
*/
STATIC SIG_CATCHER
Catcher()
{
    (void)fclose(F); (void)fclose(U); (void)close(Kmem);
    /* (void)unlink(Host); */ /* Don't want to delete on exit any more */
    exit(1);
}

int read_init()
{
    if (chdir(SPOOLDIR) < 0) {
        if (showerr)
        {
	fprintf(stderr, "%s: can't cd to \"%s\", %s...retrying every %d seconds\n",
	    progname, SPOOLDIR, strerror(errno),NAP_TIME);
        showerr=0;
        }
	return(1);
    }

    /* Open utmp. */
    if ((U = fopen("/etc/utmp", "r")) == NULL) {
	fprintf(stderr, "%s: can't read utemp, %s.\n",
	    progname, strerror(errno));
	return(1);
    }

    if ((Kmem = open("/dev/kmem", O_RDONLY)) < 0) {
        (void)fclose(U);
	fprintf(stderr, "%s: can't read kmem, %s.\n",
	    progname, strerror(errno));
	return(1);
    }

    /* Open our output file, set big buffering. */
    if ((i = open(Host, O_RDWR | O_CREAT | O_TRUNC, 0666)) < 0) {
        (void)fclose(U); (void)close(Kmem);
	fprintf(stderr, "%s: can't create \"%s\", %s.\n",
	    progname, Host, strerror(errno));
	return(1);
    }

    F = fdopen(i, "w+");
    return(0);
}

int init()
{
    if (gethostname(Host, sizeof Host) < 0) {
	fprintf(stderr, "%s: can't get hostname, %s.\n",
	    progname, strerror(errno));
	return(1);
    }

    /* Read namelist from kernel. */
    Nlist[NL_AVENRUN].n_value = 0;
    Nlist[NL_BOOTTIME].n_value = 0;
#ifdef __hpux
    if (nlist("/stand/vmunix", Nlist) < 0
#else
    if (nlist("/vmunix", Nlist) < 0
#endif
     || Nlist[NL_AVENRUN].n_value == 0
     || Nlist[NL_BOOTTIME].n_value == 0) {
	fprintf(stderr, "%s: can't read namelist, %s.\n",
		progname, strerror(errno));
	return(1);
    }

    /* Open kmem, get boot time. */
    if ((Kmem = open("/dev/kmem", O_RDONLY)) < 0) {
	fprintf(stderr, "%s: can't read kmem, %s.\n",
	    progname, strerror(errno));
	return(1);
    }
    (void)lseek(Kmem, (long)Nlist[NL_BOOTTIME].n_value, L_SET);
    (void)read(Kmem, (char *)&BootTime, sizeof BootTime);
    (void)close(Kmem);

    return(0);
}

main()
{
    struct utmp		Uentry;
    struct stat		Sb;
    char		buff[BUFFLEN];
    char		line[20];
    char		name[20];
    time_t		Now;
#ifdef	vax
    double		avenrun[3];
#else
#ifdef	sun
    int			avenrun[3];
#else
#ifdef  __hpux
    double              avenrun[3];
#endif  /* hpux */
#endif	/* sun */
    /* put other types here... */
#endif	/* vax */

    progname = "lwhod";

#ifdef	DAEMON_MUST_RUN_AS_ROOT
    if (getuid()) {
	fprintf(stderr, "%s: permission denied, %s.\n",
	    progname, strerror(errno));
	exit(1);
    }
#endif	/* DAEMON_MUST_RUN_AS_ROOT */

    /* Close stdin, stdout. */
    (void)close(0);
    (void)close(1);

#if 	!defined(DEBUG) && !defined(SABER)
    if (fork() > 0)
	exit(0);
#ifdef	TIOCNOTTY
    if ((i = open("/dev/tty", O_RDWR)) >= 0) {
	(void)ioctl(i, TIOCNOTTY, (char *)0);
	(void)close(i);
    }
#else
    setpgrp(0,0); /* Added by RKL for Series 300 */
#endif	/* TIOCNOTTY */
#endif	/* !defined(DEBUG) && !defined(SABER) */

    /* Arrange to remove file on a termination signal. */
    (void)signal(SIGHUP, Catcher);
    (void)signal(SIGINT, Catcher);
    (void)signal(SIGQUIT, Catcher);
    (void)signal(SIGALRM, Catcher);
    (void)signal(SIGTERM, Catcher);
    alarm(ALARM_TIME);
    while (init()) { alarm(ALARM_TIME); sleep(NAP_TIME); }
    alarm(ALARM_TIME);
    stalled=read_init();

    /* Enter the land of for-ever. */
    for ( ; ; ) {
        while (stalled)
        {
            alarm(ALARM_TIME);
            chdir("/");
            (void)sleep(NAP_TIME);
            stalled=read_init();
        }
	fprintf(F, "%s\n", VERSION);
	fprintf(F, "%ld # boottime\n", (long)BootTime);

	/* Do the load average. */
	(void)lseek(Kmem, (long)Nlist[NL_AVENRUN].n_value, L_SET);
	(void)read(Kmem, (char *)avenrun, sizeof avenrun);
#ifdef	vax
	for (i = 0; i < 3; i++)
	    fprintf(F, "%ld ", (long)(avenrun[i] * 100));
	fprintf(F, "# loadave vax-style\n");
#else
#ifdef	sun
	for (i = 0; i < 3; i++)
	    fprintf(F, "%ld ", (long)(avenrun[i] / (FSCALE / 100)));
	fprintf(F, "# loadave sun-style\n");
#else
#ifdef	__hpux
	for (i = 0; i < 3; i++)
	    fprintf(F, "%ld ", (long)(avenrun[i] * 100));
	fprintf(F, "# loadave hpux-style\n");
#else
	/* put other types here... */
	fprintf(F, "0 0 0 # loadave unknown\n");
#endif  /* hpux */
#endif	/* sun */
#endif	/* vax */

	/* Write utmp entries. */
	rewind(U);
	(void)time(&Now);
	while (fread((char *)&Uentry, sizeof Uentry, 1, U) == 1) {
	    if (Uentry.ut_time == 0 || Uentry.ut_name[0] == '\0')
		continue;

#if	0
#ifdef	nonuser
	    /*  List multiple entries for window-systems users? */
	    if (nonuser(Uentry))
		continue;
#endif	/* nonuser */
#endif	/* 0 */

#ifdef __hpux
            if (Uentry.ut_type!=USER_PROCESS) continue;
            if (kill(Uentry.ut_pid,0)==-1) continue;
#endif

	    /* Get user name and the line they're on. */
	    (void)strncpy(name, Uentry.ut_name, sizeof name);
	    name[8] = '\0'; /* Changed by RKL to fix bug - array not zero-termed if 8 chars username */
	    (void)strncpy(line, Uentry.ut_line, sizeof line);
	    line[sizeof Uentry.ut_line - 1] = '\0';

	    (void)sprintf(buff, "/dev/%s", line);

	    fprintf(F, "%s %s %ld %ld # tty user.line.onsince.idle\n",
		name, line, (long)Uentry.ut_time,
		stat(buff, &Sb) < 0 ? 0L : (long)Now - Sb.st_atime);
	}

	fprintf(F, "%s\n", VERSION);
        stalled=1; (void)fclose(F); (void)close(Kmem); (void)fclose(U);
        chmod(Host,0666);
    }
}
